/*
 * TismTool © Tis Veugen, 2010-2011
 * Simulation, Verification
 *
 * TsmThread.c
 * version 5
 */
#include <string.h>
#include <stdio.h>
#include <setjmp.h>
#include "OSAL_Gen.h"
#include "OSAL_Thread.h"
#include "OSAL_TriggerSem.h"
#include "Tsm.h"
#include "TsmObject.h"
#include "TsmThread.h"
#include "TsmThreadType.h"
#include "TsmExtQueue.h"
#include "TsmIntQueue.h"
#include "TsmMutex.h"
#include "TsmRzVousSem.h"
#include "TsmTriggerSem.h"
#include "TsmEnvelope.h"
#include "TsmTrigger.h"
#include "TsmRvState.h"
#include "TsmSupervisor.h"

static void TsmThread_Worker( TsmThread* oTsmThread );

TsmThread* TsmThread_Ctor( TsmThreadType Type )
{
  TsmThread* this = (TsmThread*) Tsm_Malloc( sizeof(TsmThread) );
  TsmThread_Ftor( this, Type );
  return this;
}

void TsmThread_Ftor( TsmThread* this, TsmThreadType Type )
{
  TsmObject_Ftor( (TsmObject*) this );
  this->Class = "TsmThread";
  this->mType = TsmThreadType_SUPERVISOR;
  this->mSequence = 0;
  this->mbHalt = FALSE;
  this->mThread = OSAL_HANDLE_INVALID;
  this->mRendezVousThreadBid = OSAL_HANDLE_INVALID;
  this->mName[0] = '\0';
  this->mPendingTrigger = NULL;

  this->Gtor = (Tsm_pfObj) TsmThread_Gtor;
  this->GetThread = TsmThread_GetThread;
  this->GetExtQueue = TsmThread_GetExtQueue;
  this->GetIntQueue = TsmThread_GetIntQueue;
  this->GetMaxExtQueueSize = TsmThread_GetMaxExtQueueSize;
  this->GetMaxIntQueueSize = TsmThread_GetMaxIntQueueSize;
  this->HasMessage = TsmThread_HasMessage;
  this->MaxQueueSizes = TsmThread_MaxQueueSizes;
  this->Prepare = TsmThread_Prepare;
  this->Run = TsmThread_Run;
  this->Halt = TsmThread_Halt;
  this->Interrupt = TsmThread_Interrupt;
  this->Cease = TsmThread_Cease;
  this->ExclusiveLock = TsmThread_ExclusiveLock;
  this->ExclusiveUnlock = TsmThread_ExclusiveUnlock;
  this->ExclusiveReset = TsmThread_ExclusiveReset;
  this->RendezVousWait = TsmThread_RendezVousWait;
  this->RendezVousNotify = TsmThread_RendezVousNotify;
  this->TriggerWait = TsmThread_TriggerWait;
  this->TriggerNotify = TsmThread_TriggerNotify;
  this->ASyncCall = TsmThread_ASyncCall;
  this->SyncCall = TsmThread_SyncCall;
  this->Handle = TsmThread_Handle;
  this->NativeCall = TsmThread_NativeCall;
  this->CbHandle = TsmThread_CbHandle;
  this->TsmReset1 = TsmThread_TsmReset1;
  this->TsmReset2 = TsmThread_TsmReset2;
  this->Snap = TsmThread_Snap;
  this->GetName = TsmThread_GetName;
  this->SetName = TsmThread_SetName;

  this->mType = Type;
  TsmExtQueue_Ftor( &this->mExternalQueue );
  TsmIntQueue_Ftor( &this->mInternalQueue );
  TsmMutex_Ftor( &this->mExclusiveMutex );
  TsmRzVousSem_Ftor( &this->mRendezVousSem );
  TsmTriggerSem_Ftor( &this->mTriggerSem );
}

void TsmThread_Gtor( TsmThread* this )
{
  if ( this->mThread != OSAL_HANDLE_INVALID )
  {
    OSAL_ThreadDelete( this->mThread );
  }
  this->mExternalQueue.Gtor( (TsmObject*) &this->mExternalQueue );
  this->mInternalQueue.Gtor( (TsmObject*) &this->mInternalQueue );
  TsmMutex_Gtor( &this->mExclusiveMutex );
  TsmRzVousSem_Gtor( &this->mRendezVousSem );
  TsmTriggerSem_Gtor( &this->mTriggerSem );
}

OSAL_Handle TsmThread_GetThread( TsmThread* this )
{
  return this->mThread;
}

TsmExtQueue* TsmThread_GetExtQueue( TsmThread* this )
{
  return &this->mExternalQueue;
}

TsmIntQueue* TsmThread_GetIntQueue( TsmThread* this )
{
  return &this->mInternalQueue;
}

int TsmThread_GetMaxExtQueueSize( TsmThread* this )
{
  return this->mExternalQueue.GetMaxQueueSize( (TsmQueue*) &this->mExternalQueue );
}

int TsmThread_GetMaxIntQueueSize( TsmThread* this )
{
  return this->mInternalQueue.GetMaxQueueSize( (TsmQueue*) &this->mInternalQueue );
}

OSAL_Bool TsmThread_HasMessage( TsmThread* this )
{
  OSAL_Bool bHas;

  bHas = ( this->mExternalQueue.Number( (TsmQueue*) &this->mExternalQueue )
         + this->mInternalQueue.Number( (TsmQueue*) &this->mInternalQueue ) > 0 );
  return bHas;
}

void TsmThread_MaxQueueSizes( TsmThread* this, TsmStr sout )
{
  sprintf( sout, "%3d  %3d  %s",
    this->GetMaxExtQueueSize( this ),
    this->GetMaxIntQueueSize( this ),
    this->GetName( this ) );
}

void TsmThread_Prepare( TsmThread* this, char* sName )
{
  this->mbHalt = FALSE;
  this->mThread = OSAL_ThreadCreate( (void (*)(void *)) TsmThread_Worker, this );
  if ( OSAL_HANDLE_INVALID == this->mThread )
  {
    TsmException_Quit( "TsmThread_Prepare: OSAL_ThreadCreate" );
  }
  sprintf( this->mName, "Tism%d_%s", this->mThread, sName );
  this->mSequence = TsmSupervisor_Register2( this, this->mType );
  this->mExternalQueue.SetSeqId( (TsmQueue*) &this->mExternalQueue, this->mSequence );
  this->mInternalQueue.SetSeqId( (TsmQueue*) &this->mInternalQueue, this->mSequence );

  /*
   * Force all threads running on the first cpu.
   */
  OSAL_ThreadSetAffinity( this->mThread, 1 );
}

void TsmThread_Run( TsmThread* this )
{
  this->TriggerNotify( this );
}

void TsmThread_Halt( TsmThread* this )
{
  this->mbHalt = TRUE;
  this->TriggerNotify( this );
}

void TsmThread_Interrupt( TsmThread* this )
{
  TsmThread_RendezVousNotify( this );
}

void TsmThread_Cease( TsmThread* this )
{
}

void TsmThread_ExclusiveLock( TsmThread* this )
{
  TsmMutex_WaitOne( &this->mExclusiveMutex );
}

void TsmThread_ExclusiveUnlock( TsmThread* this )
{
  TsmMutex_Release( &this->mExclusiveMutex );
}

void TsmThread_ExclusiveReset( TsmThread* this )
{
  TsmMutex_Reset( &this->mExclusiveMutex );
}

void TsmThread_RendezVousWait( TsmThread* this )
{
  this->mRendezVousThreadBid = OSAL_GetCurrentThreadHandle();
  TsmRzVousSem_WaitOne( &this->mRendezVousSem );
}

void TsmThread_RendezVousNotify( TsmThread* this )
{
  Tsm_Seq2( "R,", this->mExternalQueue.GetSeqId( (TsmQueue*) &this->mExternalQueue ) );
  TsmSupervisor_SetState( this->mRendezVousThreadBid );
  TsmRzVousSem_Release( &this->mRendezVousSem );
}

void TsmThread_TriggerWait( TsmThread* this )
{
  TsmTriggerSem_WaitOne( &this->mTriggerSem );
}

void TsmThread_TriggerNotify( TsmThread* this )
{
  TsmTriggerSem_Release( &this->mTriggerSem );
}

void TsmThread_ASyncCall( TsmThread* this, TsmEnvelope* oEnvelope )
{
  this->mExternalQueue.Put( (TsmQueue*) &this->mExternalQueue, oEnvelope );
}

void TsmThread_SyncCall( TsmThread* this, TsmEnvelope* oEnvelope )
{
  TsmTrigger* lTrigger;
  OSAL_Handle CurrentHandle;

  this->ExclusiveLock( this );

  lTrigger = TsmEnvelope_GetTrigger( oEnvelope );
  lTrigger->RvState = TsmRvState_WAIT;
  this->mPendingTrigger = lTrigger;
  this->mExternalQueue.Put( (TsmQueue*) &this->mExternalQueue, oEnvelope );

  CurrentHandle = OSAL_GetCurrentThreadHandle();
  TsmSupervisor_SetStateRv( CurrentHandle, this );
  this->RendezVousWait( this );
  this->mPendingTrigger = NULL;
  TsmSupervisor_SetState2( CurrentHandle, TsmThreadState_WAITING_RV2 );

  this->ExclusiveUnlock( this );
}

TsmRvState TsmThread_Handle( TsmThread* this, TsmEnvelope* oEnvelope )
{
  TsmTrigger* lTrigger = TsmEnvelope_GetTrigger( oEnvelope );
  TsmClass*   lTarget  = TsmEnvelope_GetTarget( oEnvelope );

  lTarget->ExecuteTrigger( lTarget, lTrigger );
  if ( lTrigger->RvState == TsmRvState_READY )
  {
    this->RendezVousNotify( this );
    lTrigger->RvState = TsmRvState_IDLE;
  }
  return lTrigger->RvState;
}

void TsmThread_NativeCall( TsmThread* this, TsmEnvelope* oEnvelope )
{
  /* To be overridden! */
}

void TsmThread_CbHandle( TsmThread* this, TsmEnvelope* oEnvelope )
{
  /* To be overridden! */
}

void TsmThread_TsmReset1( TsmThread* this )
{
  this->mExternalQueue.Empty( (TsmQueue*) &this->mExternalQueue );
  this->mInternalQueue.Empty( (TsmQueue*) &this->mInternalQueue );
}

void TsmThread_TsmReset2( TsmThread* this )
{
  TsmRzVousSem_Reset( &this->mRendezVousSem );
  this->ExclusiveReset( this );
  if ( this->mPendingTrigger != NULL )
  {
    this->mPendingTrigger->Dtor( (TsmObject*) this->mPendingTrigger );
    this->mPendingTrigger = NULL;
  }
}

void TsmThread_Snap( TsmThread* this, TsmStr sout )
{
  *sout = '\0';
  if ( this->mExternalQueue.Number( (TsmQueue*) &this->mExternalQueue ) > 0 )
  {
    strcat( sout, " Q:" );
  }
  this->mExternalQueue.Snap( (TsmQueue*) &this->mExternalQueue, sout+strlen(sout) );

  if ( this->mInternalQueue.Number( (TsmQueue*) &this->mInternalQueue ) > 0 )
  {
    strcat( sout, " S:" );
  }
  this->mInternalQueue.Snap( (TsmQueue*) &this->mInternalQueue, sout+strlen(sout) );
}

char* TsmThread_GetName( TsmThread* this )
{
  return this->mName;
}

void TsmThread_SetName( TsmThread* this, char* s )
{
  strcpy( this->mName, s );
}

static void TsmThread_Worker( TsmThread* oTsmThread )
{
  TsmEnvelope   lEnvelope;
  TsmTrigger*   lTrigger = NULL;
  TsmRvState    lRvState = TsmRvState_IDLE;
  OSAL_Thread*  oThread;
  int           rv;

  oTsmThread->TriggerWait( oTsmThread );

  oThread = OSAL_GetCurrentThread();
  rv = setjmp( oThread->JmpBuf );
  lTrigger = oTsmThread->mLocalTrigger;
  switch ( rv )
  {
  case TsmException_OK :
    break;
  case TsmException_QUIT :
    Tsm_ShowException( "Unexpected error" );
    lTrigger->DBU( lTrigger );
    break;
  case TsmException_RV :
    lRvState = TsmRvState_IDLE;
    TsmSupervisor_SetState2( oTsmThread->mThread, TsmThreadState_INTERRUPTED );
    lTrigger->DBU( lTrigger );
    break;
  }

  while ( ! oTsmThread->mbHalt )
  {
    /* Wait for message controlled by Supervisor */
    TsmSupervisor_SetState2( oTsmThread->mThread, TsmThreadState_WAITING_MSG );

    if ( oTsmThread->mInternalQueue.Dequeue
          ( (TsmQueue*) &oTsmThread->mInternalQueue, &lEnvelope ) )
    {
      lTrigger = TsmEnvelope_GetTrigger( &lEnvelope );
      lTrigger->RvState = lRvState;

      Tsm_Seq2( "I,", oTsmThread->mSequence );
      oTsmThread->mLocalTrigger = lTrigger;
      lRvState = oTsmThread->Handle( oTsmThread, &lEnvelope );
      lTrigger->DBU( lTrigger );
    }
    else
    {
      if ( oTsmThread->mExternalQueue.Dequeue
            ( (TsmQueue*) &oTsmThread->mExternalQueue, &lEnvelope ) )
      {
        lTrigger = TsmEnvelope_GetTrigger( &lEnvelope );
        if ( lTrigger->RvState == TsmRvState_WAIT )
        {
          lRvState = TsmRvState_WAIT;
        }
        else
        {
          lTrigger->RvState = lRvState;
        }

        Tsm_Seq2( "E,", oTsmThread->mSequence );
        oTsmThread->mLocalTrigger = lTrigger;
        lRvState = oTsmThread->Handle( oTsmThread, &lEnvelope );
        lTrigger->DBU( lTrigger );
      }
    }
  }
}
